#
# Cynical Behaviors for Indigo 6+
#
# Copyright 2013-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# KeypadLinc-related behaviors.
#
import time
import datetime

import asyn
import cyin.devstate
import cyin.stddevice
from cyin.stddevice import InsteonDevice
from cyin import iom, plug
from cyin import log, debug, error
from cyin.asynplugin import action
from cyin.check import *


#
# KPL button meanings
#
PRESS	= "press"				# normal button press
INSTANT = "instant"				# double-press ("instant on/off")
DIM		= "dim"					# long press (start dimming)
STOP_DIM= "stop dim"			# release after long press (stop dimming)

INSTEON_FUNCTIONS = {
#	Indigo string:			{on/off, type},
	"on":					(True, PRESS),
	"off":					(False, PRESS),
	"on to 100% (instant)":	(True,	INSTANT),
	"off to 0% (instant)":	(False,	INSTANT),
	"start brighten":		(True,	DIM),
	"stop brighten":		(True,	STOP_DIM),
	"start dim":			(False,	DIM),
	"stop dim":				(False,	STOP_DIM)
}


#
# Common behaviors of a KeypadLinc secondary button.
# The target is assumed to be a device; subclasses can override that.
#
class ButtonBehavior(cyin.Device):

	device = cyin.PluginProperty(type=cyin.device)
	button = cyin.PluginProperty(type=int)

	TARGETKIND = 'device'
	
	deferred = False

	def start(self):
		self.observe(self.TARGETKIND, [self.target])
		self.set_display_address(u"%s #%s" % (self.device.name, self.button))
		self.setLed()
	
	def _kpl_event(self, cmd):
		""" For our KPL button (only), call self.kplFunc indicating the button action. """
		if cmd.address == self.device.address:
			if cmd.cmdScene == self.button:
				func = cmd.cmdFunc
				debug(self.name, "button", self.button, "received", func)
				op = INSTEON_FUNCTIONS.get(func)
				if op:
					self.kplFunc(*op)
				else:
					error("unrecognized cmdFunc", func)
	
	def notify(self, kind, op, dev):
		""" Track changes to target device and reflect its onOff state in the button LED. """
		assert kind == self.TARGETKIND
		if UpdateLEDs.deferring():
			self.deferred = True
		else:
			self.setLed()
	
	def setLed(self):
		if self.device.canSetLed(self.button):
			on = self.targetState()
			if on != self.device.getLed(self.button):
				debug(self.name, "target changed to", on, "setting led", self.button, "to match")
				self.device.setLed(self.button, on)

	def targetState(self):
		return self.target.onOff
	
	def performDeferredUpdates(self, ctx):
		self.setLed()
		self.deferred = False
	

#
# Link a KeypadLinc button to an arbitrary on/off device.
#
INSTEON_RETRANSMIT = [1, 3, 8]		# retransmission backup strategy (cum. seconds)
INSTEON_DELAY = 0.3					# delay sending INSTEON commands after KPL notification to avoid bus jam

def insteonDelay(dev):
	return INSTEON_DELAY if isinstance(dev, InsteonDevice) else 0


class ButtonDevice(ButtonBehavior):
	""" Link a KeypadLinc button to an Indigo on/off device of any kind. """
	target = cyin.PluginProperty(type=cyin.device)

	def kplFunc(self, on, instant):
		self.setTarget(on)		
	
	def setTarget(self, on, retrans = INSTEON_RETRANSMIT):
		debug(self.name, "switching target to", on)
		if on != self.target.onOff:
			cyin.plugin.schedule(lambda ctx: self.target.turn(on), after=insteonDelay(self.target))
			if retrans:		# schedule retransmission if this one doesn't go through
				cyin.plugin.schedule(lambda ctx: self.setTarget(on, retrans[1:]), after=retrans[0])
			else:
				error(self.name, "failed to set", self.target.name, "to", on)


#
# Specifically control a Speed Control Device (aka FanLinc) with a KeypadLinc button.
# The button is expected to be in toggle mode; on and instant-on can set different speeds.
# Off is always off.
#
class ButtonFan(ButtonBehavior):
	""" Link a KeypadLinc button to a speed control (FanLinc et al) device. """
	target = cyin.PluginProperty(type=cyin.device)
	speed1 = cyin.PluginProperty(type=int)
	speed2 = cyin.PluginProperty(type=int)

	def kplFunc(self, on, op):
		if on:
			if op == INSTANT or op == DIM:
				self.setTarget(self.speed2)
			else:
				self.setTarget(self.speed1)
		else:
			self.setTarget(0)

	def setTarget(self, speed, retrans = INSTEON_RETRANSMIT):
		debug(self.name, "switching target to", speed)
		if (speed != 0) != self.target.onOff:
			cyin.plugin.schedule(lambda ctx: indigo.speedcontrol.setSpeedIndex(self.target.id, speed), after=insteonDelay(self.target))
			if retrans:
				cyin.plugin.schedule(lambda ctx: self.setTarget(speed, retrans[1:]), after=retrans[0])
			else:
				error(self.name, "failed to set", self.target.name, "to", speed)


#
# Link a button to an Indigo variable.
#
class ButtonVariable(ButtonBehavior):
	""" Link an INSTEON button to an Indigo variable. """
	target = cyin.PluginProperty(type=lambda v: cyin.variable(int(v)))	# the variable and...
	targetVar = cyin.NamedVariable(name="target", type=bool) # ... access to it

	TARGETKIND = 'variable'
	
	def kplFunc(self, on, op):
		if op in [PRESS, INSTANT]:
			self.targetVar = on

	def targetState(self):
		return self.targetVar


#
# Actions
#
class UpdateLEDs(cyin.Action):
	duration = cyin.PluginProperty(type=float, check=[check_range(min=0, max=60)])
	
	deferTimer = None

	@action
	def perform(self):
		self.defer(self.duration)
	
	@classmethod
	def deferring(cls):
		return cls.deferTimer is not None
		
	@classmethod
	def defer(cls, duration):
		""" Set or reset the update deferral timer. Duration may be zero to re-enable immediately. """
		if cls.deferTimer:
			cls.deferTimer.cancel()
		if duration == 0:
			cls.performUpdates(None)
		else:
			cls.deferTimer = cyin.plugin.schedule(cls.performUpdates, after=duration)
			debug("deferring LED updates for", duration, "seconds")
		
	@classmethod
	def performUpdates(cls, ctx):
		""" Perform deferred updates on our objects. """
		cls.deferTimer = None
		debug("updating LED states")
		delay = 0
		for iom in ButtonBehavior.all():
			if iom.deferred or ctx is None:		# ctx==None means instant action
				cyin.plugin.schedule(iom.performDeferredUpdates, after=delay)
				delay += INSTEON_DELAY


#
# Filters
#
class ButtonControllerFilter(cyin.MenuGenerator):
	""" Menu filter - KeypadLinc devices (both 6 and 8 button) """
	def generate(self):
		for io in indigo.devices.iter("indigo.insteon,indigo.controller"):
			if hasattr(io, "ledStates"):
				if len(io.ledStates) > 0:
					yield cyin.device(io.id).filter_clause()


class KPLButtonFilter(cyin.MenuGenerator):
	""" Menu filter - available buttons on a KPL device """
	def generate(self):
		device = self.ui.device
		if device:
			config = device.ledConfig
			for b in range(0, len(config)):
				if config[b]:
					yield (b+1, "Button %d" % (b+1))	# normalize to base 1
