#
# Cynical Behaviors for Indigo 6+
#
# Copyright 2013-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Presence/activity behaviors.
#
import time
import datetime

import asyn
import cyin.devstate
import cyin.stddevice
from cyin.stddevice import InsteonDevice
from cyin import iom, plug
from cyin import log, debug, error
from cyin.asynplugin import action
from cyin.check import *
	

#
# A device that tracks multiple inputs (both sensor devices and actions) and resolves
# them into a notion of presence.
#
class PresenceManager(cyin.Device):
	""" A device that models presence based on unlimited configured inputs. """
	duration = cyin.PluginProperty(type=float, check=[check_range(min=1, max=86400)], eval=True)
	condition = cyin.PluginProperty(type=lambda v: cyin.variable(int(v)), required=False, reconfigure=False)
	condition_var = cyin.NamedVariable(type=bool, name='condition', required=False)
	sensors = cyin.PluginProperty(type=lambda v: map(cyin.device, v), untype=indigo.List, required=False)
	perimeter = cyin.PluginProperty(type=lambda v: map(cyin.device, v), untype=indigo.List, required=False)
	cycletime = cyin.PluginProperty(type=float, check=[check_range(min=0, max=600)], reconfigure=False, eval=True)

	timer = None			# active turn-off Scheduled, if any
	last_off = 0			# timestamp of last turn-off performed
	
	def start(self):
		self.observe('device', [self.target] + (self.sensors or []) + (self.perimeter or []))
	
	def notify(self, kind, op, dev):
		""" Object change observations """
		turnedOn = dev.onOff and dev.stateChanged("onOffState")
		debug(self.name, "notify", kind, op, "device", dev, "is", dev.onOff, "edge", turnedOn)
		if dev == self.target:
			if turnedOn:
				self.setTimer(self.timerBase(), self.duration)
			else:
				self.cancelTimer()
				self.last_off = time.time()
		elif self.condition and self.condition_var:
			if dev.onOff:
				debug(self.name, "ignoring", dev.name, "because variable", self.condition.name, "is true")
		elif self.sensors and dev in self.sensors:
			if dev.onOff:
				self._trigger(False)
		elif self.perimeter and dev in self.perimeter:
			if turnedOn:
				self._trigger(True)
		else:
			error(self.name, "unexpected device notification for", dev.name)

	def actionTrigger(self, perimeter, override):
		""" Process a trigger by explicit action """
		if not override and self.condition and self.condition_var:
			# blocked by variable and not overridden
			debug(self.name, "ignoring explicit trigger because variable", self.condition.name, "is true")
		else:
			self._trigger(perimeter)
	
	def _trigger(self, perimeter):
		""" Process a secondary trigger event of any kind """
		if self.last_off + self.cycletime > time.time():
			debug(self.name, "cycletime suppresses sensor trigger")
			return
		if self.target.onOff:
			if not perimeter:
				self.setTimer(time.time(), self.duration)
		else:
			self.target.turn(True)
	
	def timerBase(self):
		return self.timer.when if self.timer else time.time()
	
	def setTimer(self, base, offset):
		timeout = base + offset
		if self.timer and self.timer.when > timeout:
			return
		if self.timer:
			debug(self.name, "resetting timer")
			self.timer.cancel()
		self.timer = cyin.plugin.schedule(self.timeout, at=timeout)
		debug(self.name, "timeout now", self.timer.when - time.time())
	
	def cancelTimer(self):
		if self.timer:
			debug(self.name, "cancel timer")
			self.timer.cancel()
			self.timer = None
	
	def timeout(self, ctx):
		debug(self.name, "timeout - turning off", self.target.name)
		self.timer = None
		self.target.turn(False)


#
# A PresenceManager that is its own relay device.
# The onOff state represents presence.
# 
class Presence(PresenceManager):
	""" A device that represents presence as its own relay device onOff state. """

	def __init__(self, io):
		super(Presence, self).__init__(io)
		self.target = self
	
	@action
	def standard_switch(self, active):
		self.onOff = active
	
	@action
	def standard_toggle(self, active):
		self.onOff = not self.onOff


#
# A PresenceManager that switches an explicitly given relay device.
# The target device will be switched on for presence and off for absence.
#
class SwitchControl(PresenceManager):
	""" A device that switches a target device based on presence. """
	target = cyin.PluginProperty(type=cyin.device)
	
	def start(self):
		super(SwitchControl, self).start()
		self.set_display_address(self.target.name)


#
# Actions
#
class TriggerSwitchControl(cyin.DeviceAction):
	""" Explicitly trigger a Switch Control device. """
	perimeter = cyin.PluginProperty(type=bool)
	override = cyin.PluginProperty(type=bool)
	
	@action
	def perform(self):
		self.device.actionTrigger(self.perimeter, self.override)


#
# Filters
#
class BinarySensorFilter(cyin.MenuGenerator):
	""" Menu filter - heuristic binary sensor device filter.
	
		This filter produces genuine indigo.sensor devices that have
		onOffState state. It also yields INSTEON I/O-Linc devices
		which cyin internally conforms to the pattern.
		
		If the loosedevices preference is true, we show ANY device that
		has an onOffState. This catches custom devices that should be sensors
		but are not, at the price of listing a bunch of useless devices as well.
	"""
	def generate(self):
		sensors = [io.id for io in indigo.devices.iter("indigo.sensor")]
		loose = cyin.plugin.loosedevices
		for io in indigo.devices.iter():
			# special deal for I/O-Linc devices (cyin makes them look like sensors)
			if io.protocol == indigo.kProtocol.Insteon and io.model.startswith("I/O-Linc"):
				yield cyin.device(io.id).filter_clause()
			if "onOffState" in io.states:
				if loose or io.id in sensors:
					yield cyin.device(io.id).filter_clause()
